/*
 * Velleman K8048 Programmer for FreeBSD and others.
 *
 * Copyright (c) 2005-2007 Darron Broad
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE
 * for details.
 *
 * $Id: inhx32.c,v 1.31 2008/02/29 17:37:05 darron Exp $
 */

#include "k8048.h"

/*
 * Get hex nibble
 */
unsigned char
inhx32_gethexn(char c)
{
    if(c>='0' && c<='9')
        return c-'0';
    if(c>='a' && c<='f')
        return c-'a'+10;
    if(c>='A' && c<='F')
        return c-'A'+10;
    return 0;
}

/*
 * Get hex byte
 */
unsigned char
inhx32_gethexb(char *s)
{
    if(strlen(s)>1)
        return inhx32_gethexn(s[0]) << 4 | inhx32_gethexn(s[1]);
    return 0;
}

/*
 * Compare inhx32 data
 */
int
inhx32_compare(const void *d1, const void *d2)
{
    inhx32_data *data1= (inhx32_data *)d1;
    inhx32_data *data2= (inhx32_data *)d2;

    if(data1->address < data2->address)
        return -1;
    else if(data1->address == data2->address)
        return 0;
    return 1;
}

/*
 * Create ordered list of inhx32 data
 */
void 
inhx32_order(const void *entry, VISIT order, int level) 
{ 
    inhx32_data *inhx32= *(inhx32_data **)entry;

    if(order==postorder || order==leaf) 
        inhx32_pdata[inhx32_index++]= inhx32;
} 

/*
 * inhx32 format parser
 *
 * returns the number of ordered lines in the inhx32_pdata array after parsing filename
 */
int
inhx32(char *filename)
{
    FILE *f1;
    char *line;
    unsigned int extended_address=0, bb, ix;
    int n;
    unsigned char tt=TT_DATA, cc;
    unsigned short aaaa;
    inhx32_data *root=NULL, *inhx32;

    inhx32_count=0;
    inhx32_pdata=NULL;
    
    f1=fopen(filename, "rb");
    if(f1==NULL)
    {
        printf("%s: error: file open failed [%s] [%s]\n", __FUNCTION__, filename, strerror(errno));
        return -1;
    }
    line= (char *)calloc(STRLEN, sizeof(char));
    if(line==NULL)
    {
        printf("%s: fatal error: calloc failed\n", __FUNCTION__);
        exit(-1); /* panic */
    }

    while(tt!=TT_EOF && fgets(line, STRLEN, f1)!=NULL)
    {
        line[STRMAX]='\0';

        /* strip CRLF */
        n= strlen(line)-1;
        while(n>=0 && (line[n]=='\n' || line[n]=='\r'))
		line[n--]='\0';

        /* validate line prefix and length */
        if(line[0]!=':' || (strlen(line)&1)==0 || strlen(line)<11)
        {
            printf("%s: warning: ignoring malformed line: invalid format [%s]\n", __FUNCTION__, line);
            continue;
        }

        /* validate checksum */
        cc=0;
        for(n=1; line[n]; n+=2)
            cc+=inhx32_gethexb(&line[n]); 
        if(cc!=0)
        {
            printf("%s: warning: ignoring malformed line: invalid checksum [%d]\n", __FUNCTION__, cc);
            continue;
        }

        /* determine number of data bytes in this line */
        bb= inhx32_gethexb(&line[BB]);

        /* validate line length */
        if(strlen(line)!=(2*bb+11))
        {
            printf("%s: warning: ignoring malformed line: invalid length [%s]\n", __FUNCTION__, line);
            continue;
        }

        /* determine data address for this line */
        aaaa= (inhx32_gethexb(&line[AAAA]) << 8) | inhx32_gethexb(&line[AAAA+2]);

        /* determine record type */
        tt= inhx32_gethexb(&line[TT]);

        switch(tt)
        {
            case TT_DATA:   if(bb==0)
                            {
                                printf("%s: warning: ignoring empty line\n", __FUNCTION__);
                                break;
                            }

                            /* allocate new array entry for this line */
                            inhx32= (inhx32_data *)calloc(1, sizeof(inhx32_data));
                            if(inhx32==NULL)
                            {
                                printf("%s: fatal error: calloc failed\n", __FUNCTION__);
                                exit(-1); /* panic */
                            }

                            /* increment line counter */
                            inhx32_count++;

                            /* save address and word count */
                            inhx32->address= extended_address | aaaa;
                            inhx32->nbytes = bb;

                            /* extract data */
                            ix=HHHH;
                            for(n=0; n<bb; n++)
                            {
                                inhx32->bytes[n]= inhx32_gethexb(&line[ix]);
                                ix+=2;
                            }

                            /* add entry into tree */
                            if(tsearch((void *)inhx32, (void **)&root, inhx32_compare)==NULL)
                            {
                                printf("%s: fatal error: tsearch failed\n", __FUNCTION__);
                                exit(-1); /* panic */
                            }
                            break;

            case TT_EOF:    break;

            case TT_EXTENDED_SEGMENT_ADDRESS:
                            printf("%s: warning: ignoring unhandled extended segment address\n", __FUNCTION__);
                            break;

            case TT_START_SEGMENT_ADDRESS:
                            printf("%s: warning: ignoring unhandled start segment address\n", __FUNCTION__);
                            break;

            case TT_EXTENDED_LINEAR_ADDRESS:
                            if(aaaa==0 && bb==2)
                                extended_address= (inhx32_gethexb(&line[HHHH]) << 24) | (inhx32_gethexb(&line[HHHH+2]) << 16);
                            else
                                printf("%s: warning: ignoring invalid extended linear address [aaaa=%04x, bb=%d]\n", __FUNCTION__, aaaa, bb);
                            break;

            case TT_START_LINEAR_ADDRESS:
                            printf("%s: warning: ignoring unhandled start linear address\n", __FUNCTION__);
                            break;

            default:        printf("%s: warning: ignoring unknown record type [%d]\n", __FUNCTION__, tt);
                            break;
        }
    }
    free(line);
    fclose(f1);

    /* return error if no data lines found */
    if(inhx32_count==0)
    {
        printf("%s: error: file contains no data records [%s]\n", __FUNCTION__, filename);
        return -5;
    }
    
    /* create and populate line array from tree */
    inhx32_pdata= (inhx32_data **)calloc(inhx32_count, sizeof(inhx32_data *));
    if(inhx32_pdata==NULL)
    {
        printf("%s: fatal error: calloc failed\n", __FUNCTION__);
        exit(-1); /* panic */
    }
    inhx32_index=0;
    twalk((void *)root, inhx32_order); 

    /* free tree nodes */
    for(n=0; n<inhx32_count; n++)
    {
        if(tdelete((void *)inhx32_pdata[n], (void **)&root, inhx32_compare)==NULL)
        {
            printf("%s: fatal error: tdelete failed\n", __FUNCTION__);
            exit(-1); /* panic */
        }
    }
    if(root!=NULL)
    {
        printf("%s: fatal error: tree destroy failed\n", __FUNCTION__);
        exit(-1); /* panic */
    }

    /* return the line count */
    return inhx32_count;
}
        
/*
 * inhx32
 *
 *  free memory used by line array
 */
void
inhx32_free()
{
    if(inhx32_count)
    {
        int i;
        for(i=0; i<inhx32_count; i++)
            free(inhx32_pdata[i]);

        free(inhx32_pdata);
        
        inhx32_count=0;
        inhx32_pdata=NULL;
    }
    else
    {
        printf("%s: warning: nothing to free\n", __FUNCTION__);
    }
}
