/*
 * Velleman K8048 Programmer for FreeBSD and others.
 *
 * Copyright (c) 2005-2007 Darron Broad
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE
 * for details.
 *
 * $Id: pic18f.c,v 1.78 2008/09/19 04:42:20 darron Exp $
 *
 * PIC18 16 bit word architecture command set (all PIC18F devices should be compatible)
 *
 * PIC18F252 sample device:
 *
 *  [200000] [IDLOCATION1] ff
 *  [200001] [IDLOCATION2] ff
 *  [200002] [IDLOCATION3] ff
 *  [200003] [IDLOCATION4] ff
 *  [200004] [IDLOCATION5] ff
 *  [200005] [IDLOCATION6] ff
 *  [200006] [IDLOCATION7] ff
 *  [200007] [IDLOCATION8] ff
 *  [3FFFFE] [DEVICEID]    0407 DEV:400 REV:7 PIC18F252
 *
 *  [300000] [CONFIG1] 2700
 *  [300002] [CONFIG2] 0f0f
 *  [300004] [CONFIG3] 0100
 *  [300006] [CONFIG4] 0085
 *  [300008] [CONFIG5] c00f
 *  [30000A] [CONFIG6] e00f
 *  [30000C] [CONFIG7] 400f
 *
 * PIC18F1320 sample device:
 *
 *  [200000] [IDLOCATION1] ff
 *  [200001] [IDLOCATION2] ff
 *  [200002] [IDLOCATION3] ff
 *  [200003] [IDLOCATION4] ff
 *  [200004] [IDLOCATION5] ff
 *  [200005] [IDLOCATION6] ff
 *  [200006] [IDLOCATION7] ff
 *  [200007] [IDLOCATION8] ff
 *  [3FFFFE] [DEVICEID]    07c4 DEV:7c0 (7:6) REV:4 PIC18F1320
 *
 * PRE-BLANKING
 *  [300000] [CONFIG1] 0400
 *  [300002] [CONFIG2] 1f0f
 *  [300004] [CONFIG3] 8000
 *  [300006] [CONFIG4] 0085
 *  [300008] [CONFIG5] c003
 *  [30000A] [CONFIG6] e003
 *  [30000C] [CONFIG7] 4003
 *
 * POST-BLANKING
 *  [300000] [CONFIG1] cf00
 *  [300002] [CONFIG2] 1f0f
 *  [300004] [CONFIG3] 8000
 *  [300006] [CONFIG4] 0085
 *  [300008] [CONFIG5] c003
 *  [30000A] [CONFIG6] e003
 *  [30000C] [CONFIG7] 4003
 *
 * PIC18F2550 sample device:
 *
 *  [200000] [IDLOCATION1] ff
 *  [200001] [IDLOCATION2] ff
 *  [200002] [IDLOCATION3] ff
 *  [200003] [IDLOCATION4] ff
 *  [200004] [IDLOCATION5] ff
 *  [200005] [IDLOCATION6] ff
 *  [200006] [IDLOCATION7] ff
 *  [200007] [IDLOCATION8] ff
 *  [3FFFFE] [DEVICEID]    1242 DEV:1240 (12:2) REV:2 PIC18F2550
 *
 *  [300000] [CONFIG1] 0400
 *  [300002] [CONFIG2] 0001
 *  [300004] [CONFIG3] 8100
 *  [300006] [CONFIG4] 00c4
 *  [300008] [CONFIG5] c00f
 *  [30000A] [CONFIG6] e00f
 *  [30000C] [CONFIG7] 400f
 *
 * Compatible ICSP devices:
 *
 * DS39576B
 *  PIC18F242 U
 *  PIC18F248 U
 *  PIC18F252 T
 *  PIC18F258 U
 *  PIC18F442 U
 *  PIC18F448 U
 *  PIC18F452 U
 *  PIC18F458 U
 *
 * DS39592E
 *  PIC18F1220 U
 *  PIC18F1320 T
 *  PIC18F2220 U
 *  PIC18F2320 U
 *  PIC18F4220 U
 *  PIC18F4320 U
 *
 * DS39622K
 *  PIC18F2221 U
 *  PIC18F2321 U
 *  PIC18F2410 U
 *  PIC18F2420 U
 *  PIC18F2423 U
 *  PIC18F2450 U
 *  PIC18F2455 U
 *  PIC18F2458 U
 *  PIC18F2480 U
 *  PIC18F2510 U
 *  PIC18F2515 U
 *  PIC18F2520 U
 *  PIC18F2523 U
 *  PIC18F2525 U
 *  PIC18F2550 T
 *  PIC18F2553 U
 *  PIC18F2580 U
 *  PIC18F2585 U
 *  PIC18F2610 U
 *  PIC18F2620 U
 *  PIC18F2680 U
 *  PIC18F2682 U
 *  PIC18F2685 U
 *  PIC18F4221 U
 *  PIC18F4321 U
 *  PIC18F4410 U
 *  PIC18F4420 U
 *  PIC18F4423 U
 *  PIC18F4450 U
 *  PIC18F4455 U
 *  PIC18F4458 U
 *  PIC18F4480 U
 *  PIC18F4510 U
 *  PIC18F4515 U
 *  PIC18F4520 U
 *  PIC18F4523 U
 *  PIC18F4525 U
 *  PIC18F4550 U
 *  PIC18F4553 U
 *  PIC18F4580 U
 *  PIC18F4585 U
 *  PIC18F4610 U
 *  PIC18F4620 U
 *  PIC18F4680 U
 *  PIC18F4682 U
 *  PIC18F4685 U
 *
 * DS39752A
 *  PIC18F1230      U
 *  PIC18F1330      U
 *  PIC18F1330-ICD  U
 *
 * T=TESTED
 * U=UNTESTED
 */

#include "k8048.h"

/*****************************************************************************
 *
 * Hardware configuration
 *
 *****************************************************************************/
struct pic18f_config pic18f_conf;

/*****************************************************************************
 *
 * Hardware alogithm map
 *
 *****************************************************************************/
struct pic18f_dsmap pic18f_map[]=
{
    /*  devicename      deviceid    datasheet   revision    panelsize */
    {   "PIC18F242",    PIC18F242,  DS39576B,   0xff,       8           },
    {   "PIC18F248",    PIC18F248,  DS39576B,   0xff,       8           },
    {   "PIC18F252",    PIC18F252,  DS39576B,   0xff,       8           },
    {   "PIC18F258",    PIC18F258,  DS39576B,   0xff,       8           },
    {   "PIC18F442",    PIC18F442,  DS39576B,   0xff,       8           },
    {   "PIC18F448",    PIC18F448,  DS39576B,   0xff,       8           },
    {   "PIC18F452",    PIC18F452,  DS39576B,   0xff,       8           },
    {   "PIC18F458",    PIC18F458,  DS39576B,   0xff,       8           },

    {   "PIC18F1220",   PIC18F1220, DS39592E,   0xff,       8           },
    {   "PIC18F1320",   PIC18F1320, DS39592E,   0xff,       8           },
    {   "PIC18F2220",   PIC18F2220, DS39592E,   0xff,       8           },
    {   "PIC18F2320",   PIC18F2320, DS39592E,   0xff,       8           },
    {   "PIC18F4220",   PIC18F4220, DS39592E,   0xff,       8           },
    {   "PIC18F4320",   PIC18F4320, DS39592E,   0xff,       8           },

    {   "PIC18F2221",   PIC18F2221, DS39622K,   0xff,       8           },
    {   "PIC18F2321",   PIC18F2321, DS39622K,   0xff,       8           },
    
    {   "PIC18F2410",   PIC18F2410, DS39622K,   0xff,       32          },
    {   "PIC18F2420",   PIC18F2420, DS39622K,   0,          32          },
    {   "PIC18F2423",   PIC18F2420, DS39622K,   1,          32          },

    {   "PIC18F2450",   PIC18F2450, DS39622K,   0xff,       8           },

    {   "PIC18F2455",   PIC18F2455, DS39622K,   0xff,       32          },
    {   "PIC18F2458",   PIC18F2458, DS39622K,   0xff,       32          },
    {   "PIC18F2480",   PIC18F2480, DS39622K,   0xff,       32          },
    {   "PIC18F2510",   PIC18F2510, DS39622K,   0xff,       32          },

    {   "PIC18F2515",   PIC18F2515, DS39622K,   0xff,       64          },

    {   "PIC18F2520",   PIC18F2520, DS39622K,   0,          32          },
    {   "PIC18F2523",   PIC18F2520, DS39622K,   1,          32          },

    {   "PIC18F2525",   PIC18F2525, DS39622K,   0xff,       64          },

    {   "PIC18F2550",   PIC18F2550, DS39622K,   0xff,       32          },
    {   "PIC18F2553",   PIC18F2553, DS39622K,   0xff,       32          },
    {   "PIC18F2580",   PIC18F2580, DS39622K,   0xff,       32          },

    {   "PIC18F2585",   PIC18F2585, DS39622K,   0xff,       64          },
    {   "PIC18F2610",   PIC18F2610, DS39622K,   0xff,       64          },
    {   "PIC18F2620",   PIC18F2620, DS39622K,   0xff,       64          },
    {   "PIC18F2680",   PIC18F2680, DS39622K,   0xff,       64          },
    {   "PIC18F2682",   PIC18F2682, DS39622K,   0xff,       64          },
    {   "PIC18F2685",   PIC18F2685, DS39622K,   0xff,       64          },

    {   "PIC18F4221",   PIC18F4221, DS39622K,   0xff,       8           },
    {   "PIC18F4321",   PIC18F4321, DS39622K,   0xff,       8           },

    {   "PIC18F4410",   PIC18F4410, DS39622K,   0xff,       32          },
    {   "PIC18F4420",   PIC18F4420, DS39622K,   0,          32          },
    {   "PIC18F4423",   PIC18F4420, DS39622K,   1,          32          },

    {   "PIC18F4450",   PIC18F4450, DS39622K,   0xff,       8           },

    {   "PIC18F4455",   PIC18F4455, DS39622K,   0xff,       32          },
    {   "PIC18F4458",   PIC18F4458, DS39622K,   0xff,       32          },
    {   "PIC18F4480",   PIC18F4480, DS39622K,   0xff,       32          },
    {   "PIC18F4510",   PIC18F4510, DS39622K,   0xff,       32          },

    {   "PIC18F4515",   PIC18F4515, DS39622K,   0xff,       64          },

    {   "PIC18F4520",   PIC18F4520, DS39622K,   0,          32          },
    {   "PIC18F4523",   PIC18F4520, DS39622K,   1,          32          },

    {   "PIC18F4525",   PIC18F4525, DS39622K,   0xff,       64          },

    {   "PIC18F4550",   PIC18F4550, DS39622K,   0xff,       32          },
    {   "PIC18F4553",   PIC18F4553, DS39622K,   0xff,       32          },
    {   "PIC18F4580",   PIC18F4580, DS39622K,   0xff,       32          },

    {   "PIC18F4585",   PIC18F4585, DS39622K,   0xff,       64          },
    {   "PIC18F4610",   PIC18F4610, DS39622K,   0xff,       64          },
    {   "PIC18F4620",   PIC18F4620, DS39622K,   0xff,       64          },
    {   "PIC18F4680",   PIC18F4680, DS39622K,   0xff,       64          },
    {   "PIC18F4682",   PIC18F4682, DS39622K,   0xff,       64          },
    {   "PIC18F4685",   PIC18F4685, DS39622K,   0xff,       64          },

    {   "PIC18F1230",   PIC18F1230, DS39752A,   0xff,       8           },
    {   "PIC18F1330",   PIC18F1330, DS39752A,   0xff,       8           },
    {   "PIC18F1330-ICD",PIC18F1330ICD,DS39752A,0xff,       8           },

    {   "(null)",       0,          0,          0,          0           }
};

int pic18f_index= (sizeof(pic18f_map) / sizeof(struct pic18f_dsmap)) - 1;

/*****************************************************************************
 *
 * Hardware functions
 *
 *****************************************************************************/

/*
 * 0000
 *
 * CORE INSTRUCTION (SHIFT IN 16 BIT INSTRUCTION)
 */
void
pic18f_core_instruction(struct k8048 *k, unsigned short word)
{
    io_command_out(k, "0000");
    io_word_out(k, word);
}

/*
 * 0000
 *
 * CORE INSTRUCTION (PROGRAMMING NOP) DS39576B-page 13
 */
void
pic18f_core_instruction_nopp(struct k8048 *k)
{
    io_command_out(k, "000");       /* 000 */
    io_set_pgd_pgc(k, LOW, HIGH);   /* 0   */
    usleep(5000);                   /* P9  */
    io_set_pgc(k, LOW);
    usleep(1);                      /* P10 */
    io_word_out(k,0);
}

/*
 * 0000
 *
 * CORE INSTRUCTION (ERASE NOP) DS39576B-page 7
 */
void
pic18f_core_instruction_nope(struct k8048 *k)
{
    io_command_out(k, "0000");
    usleep(5000);                   /* P11  */
    usleep(1);                      /* P10 */
    io_word_out(k,0);
}

/*
 * 0010
 *
 * SHIFT OUT TABLAT REGISTER
 */
unsigned char
pic18f_shift_out_tablat_register(struct k8048 *k)
{
    io_command_out(k, "0010");
    return io_byte_in(k);
}

/*
 * 1000
 *
 * TABLE READ
 */
unsigned char
pic18f_table_read(struct k8048 *k)
{
    io_command_out(k, "1000");
    return io_byte_in(k);
}

/*
 * 1001
 *
 * TABLE READ POST INCREMENT
 */
unsigned char
pic18f_table_read_post_increment(struct k8048 *k)
{
    io_command_out(k, "1001");
    return io_byte_in(k);
}

/*
 * 1010
 *
 * TABLE READ POST DECREMENT
 */
unsigned char
pic18f_table_read_post_decrement(struct k8048 *k)
{
    io_command_out(k, "1010");
    return io_byte_in(k);
}

/*
 * 1011
 *
 * TABLE READ PRE INCREMENT
 */
unsigned char
pic18f_table_read_pre_increment(struct k8048 *k)
{
    io_command_out(k, "1011");
    return io_byte_in(k);
}

/*
 * 1100
 *
 * TABLE WRITE
 */
void
pic18f_table_write(struct k8048 *k, unsigned short word)
{
    io_command_out(k, "1100");
    io_word_out(k, word);
}

/*
 * 1101
 *
 * TABLE WRITE POST INCREMENT BY 2
 */
void
pic18f_table_write_post_increment_2(struct k8048 *k, unsigned short word)
{
    io_command_out(k, "1101");
    io_word_out(k, word);
}

/*
 * 1110
 *
 * TABLE WRITE POST DECREMENT BY 2
 */
void
pic18f_table_write_post_decrement_2(struct k8048 *k, unsigned short word)
{
    io_command_out(k, "1110");
    io_word_out(k, word);
}

/*
 * 1111
 *
 * TABLE WRITE START PROGRAMMING
 */
void
pic18f_table_write_start_programming(struct k8048 *k, unsigned short word)
{
    io_command_out(k, "1111");
    io_word_out(k, word);
}

/*****************************************************************************
 *
 * Compound functions
 *
 *****************************************************************************/

/*
 * INIT CODE MEMORY ACCESS
 */
void
pic18f_init_code_memory_access(struct k8048 *k)
{
    pic18f_core_instruction(k, 0x8EA6);            /* BSF EECON1, EEPGD    */
    pic18f_core_instruction(k, 0x9CA6);            /* BCF EECON1, CFGS     */
}

/*
 * INIT CONFIG MEMORY ACCESS
 */
void
pic18f_init_config_memory_access(struct k8048 *k)
{
    pic18f_core_instruction(k, 0x8EA6);            /* BSF EECON1, EEPGD    */
    pic18f_core_instruction(k, 0x8CA6);            /* BSF EECON1, CFGS     */
}

/*
 * INIT DATA MEMORY ACCESS (EEPROM)
 */
void
pic18f_init_data_memory_access(struct k8048 *k)
{
    pic18f_core_instruction(k, 0x9EA6);            /* BCF EECON1, EEPGD    */
    pic18f_core_instruction(k, 0x9CA6);            /* BCF EECON1, CFGS     */
}

/*
 * SET TABLE POINTER
 */
void
pic18f_set_table_pointer(struct k8048 *k, unsigned long address)
{
    unsigned char addrl=  (address & 0x000000ff);       /*  7:0             */
    unsigned char addrh= ((address & 0x0000ff00) >> 8); /* 15:8             */
    unsigned char addru= ((address & 0x003f0000) >> 16);/* 21:16            */
    pic18f_core_instruction(k, 0x0E00 | addru);     /* MOVLW <Addr[21:16]>  */
    pic18f_core_instruction(k, 0x6EF8);             /* MOVWF TBLPTRU        */
    pic18f_core_instruction(k, 0x0E00 | addrh);     /* MOVLW <Addr[15:8]>   */
    pic18f_core_instruction(k, 0x6EF7);             /* MOVWF TBLPTRH        */
    pic18f_core_instruction(k, 0x0E00 | addrl);     /* MOVLW <Addr[7:0]>    */
    pic18f_core_instruction(k, 0x6EF6);             /* MOVWF TBLPTRL        */
}

/*
 * BULK ERASE
 */
void
pic18f_bulk_erase(struct k8048 *k)
{    
    switch(pic18f_map[pic18f_index].datasheet)      /* erase */
    {
        /* case DS30500A: */
        case DS39576B:  
        case DS39592E:  pic18f_init_config_memory_access(k);           /* SWITCH TO CONFIG     */
                        pic18f_set_table_pointer(k, 0x3C0004);         /* BULK ERASE CONFIG &  */
                        pic18f_table_write(k, 0x0080);                 /* BULK ERASE CHIP      */
                        pic18f_core_instruction(k, 0x0000);            /* NOP                  */
                        pic18f_core_instruction_nope(k);               /* NOP ERASE            */
                        pic18f_init_code_memory_access(k);             /* SWITCH TO CODE       */
                        break;

        case DS39622K:  pic18f_init_config_memory_access(k);           /* SWITCH TO CONFIG     */
                        pic18f_set_table_pointer(k, 0x3C0005);         /* BULK ERASE CONFIG &  */
                        pic18f_table_write(k, 0x3f3f);                 /* BULK ERASE CHIP      */
                        pic18f_set_table_pointer(k, 0x3C0004);         /* BULK ERASE CONFIG &  */
                        pic18f_table_write(k, 0x8f8f);                 /* BULK ERASE CHIP      */
                        pic18f_core_instruction(k, 0x0000);            /* NOP                  */
                        pic18f_core_instruction_nope(k);               /* NOP ERASE            */

#if 0 /* XXX may be required */
                        pic18f_set_table_pointer(k, 0x3C0005);         /* BULK ERASE CONFIG &  */
                        pic18f_table_write(k, 0x0000);                 /* BULK ERASE EEPROM    */
                        pic18f_set_table_pointer(k, 0x3C0004);         /* BULK ERASE CONFIG &  */
                        pic18f_table_write(k, 0x0084);                 /* BULK ERASE EEPROM    */
                        pic18f_core_instruction(k, 0x0000);            /* NOP                  */
                        pic18f_core_instruction_nope(k);               /* NOP ERASE            */
#endif
                        pic18f_init_code_memory_access(k);             /* SWITCH TO CODE       */
                        break;

        case DS39752A:  pic18f_init_config_memory_access(k);           /* SWITCH TO CONFIG     */
                        pic18f_set_table_pointer(k, 0x3C0005);         /* BULK ERASE CONFIG &  */
                        pic18f_table_write(k, 0x0f0f);                 /* BULK ERASE CHIP      */
                        pic18f_set_table_pointer(k, 0x3C0004);         /* BULK ERASE CONFIG &  */
                        pic18f_table_write(k, 0x8787);                 /* BULK ERASE CHIP      */
                        pic18f_core_instruction(k, 0x0000);            /* NOP                  */
                        pic18f_core_instruction_nope(k);               /* NOP ERASE            */
                        pic18f_init_code_memory_access(k);             /* SWITCH TO CODE       */
                        break;
    }
}

/*****************************************************************************
 *
 * Read block data
 *
 *****************************************************************************/

/*
 * READ CONFIGURATION MEMORY 200000 .. 200007, 300000 .. 30000d, 3ffffe .. 3fffff
 */
void
pic18f_read_config_memory(struct k8048 *k)
{
    unsigned short deviceid, revision, error;
    int i;

    io_init_program_verify(k);
    
    pic18f_init_code_memory_access(k);
    
    pic18f_set_table_pointer(k, PIC18F_ID_LOW);
    for(i=0; i<8; i++)
        pic18f_conf.index[i]= pic18f_table_read_post_increment(k);
    
    pic18f_set_table_pointer(k, PIC18F_CONFIG_LOW);
    for(i=0; i<14; i++)
        pic18f_conf.index[i + 8]= pic18f_table_read_post_increment(k);

    pic18f_set_table_pointer(k, PIC18F_DEVICEID_LOW);
    for(i=0; i<2; i++)
        pic18f_conf.index[i + 22]= pic18f_table_read_post_increment(k);

    io_standby(k);

    if(k->devicename[0])    /* Override device probe */
    {
        printf("%s: fatal error: device override not supported on this architecture (yet).\n", __FUNCTION__);
        exit(-1); /* panic */
    }

    deviceid= (pic18f_conf.index[PIC18F_CONFIG_DEVICEID2] << 8) |
              (pic18f_conf.index[PIC18F_CONFIG_DEVICEID1] & PIC18F_DEVICEID1_IDMASK);

    revision=  pic18f_conf.index[PIC18F_CONFIG_DEVICEID1] & PIC18F_DEVICEID1_REVMASK;

    i=0;
    while( pic18f_map[i].deviceid != 0)
    {
        if(pic18f_map[i].revision==0xff)    /* any deviceid match */
        {
            if(pic18f_map[i].deviceid==deviceid)
                break;
        }
        else                                /* deviceid + revision match */
        {
            if(pic18f_map[i].deviceid==deviceid && pic18f_map[i].revision==revision)
                break;
        }
        i++;
    }

    if(pic18f_map[i].deviceid == 0)
    {
        /*
         * k8048 NO POWER               ffe0 rev 31 [ffff]
         * k8048 SWITCH IN STANDBY      0000 rev  0 [0000]
         * k0848 SWITCH IN RUN          ffe0 rev 31 [ffff]
         * k0848 SWITCH IN PROG         XXXX
         */
        error= (pic18f_conf.index[PIC18F_CONFIG_DEVICEID2] << 8) | pic18f_conf.index[PIC18F_CONFIG_DEVICEID1];
        switch( error )
        {
            case 0x0000:    printf("%s: fatal error: switch is on standby, or device is not a PIC18F.\n", __FUNCTION__);
                            break;
            case 0xffff:    printf("%s: fatal error: switch is on run, or no power is supplied, or no device is installed.\n", __FUNCTION__);
                            break;
            default:        printf("%s: fatal error: unknown device: %04x rev %2d [%04x]\n", __FUNCTION__,
                                deviceid,
                                revision,
                                error);
                            break;
        }
        exit(-1); /* panic */
    }
    pic18f_index= i;
}

/*
 * READ PROGRAM MEMORY BLOCK 0000 .. 0000 + SIZE
 *
 *  RETURN BASE ADDRESS
 */
unsigned long
pic18f_read_program_memory_block(struct k8048 *k, unsigned short *data, int max)
{
    int i;

    io_init_program_verify(k);

    pic18f_init_code_memory_access(k);

    pic18f_set_table_pointer(k, PIC18F_CODE_LOW);
        
    for(i=0; i<max; i++)
    {
        data[i] = pic18f_table_read_post_increment(k);
        data[i]|= pic18f_table_read_post_increment(k) << 8;
    }
    
    io_standby(k);

    return PIC18F_BOOTBLOCK_LOW;
}

/*
 * READ DATA (EEPROM) MEMORY BLOCK 0000 .. 0000 + SIZE
 *
 *  RETURN BASE ADDRESS
 */
unsigned long
pic18f_read_data_memory_block(struct k8048 *k, unsigned char *data, int max)
{
    int i;

    io_init_program_verify(k);

    pic18f_init_data_memory_access(k);

    for(i=0; i<max; i++)
    {
        pic18f_set_data_pointer(k, i);
        data[i]= pic18f_read_data_memory(k);
    }
    
    io_standby(k);

    return PIC18F_DATA_LOW;
}

/*****************************************************************************
 *
 * Programming: Data (EEPROM)
 *
 *****************************************************************************/

/*
 * SET DATA POINTER (EEPROM)
 */
void
pic18f_set_data_pointer(struct k8048 *k, unsigned short address)
{
    unsigned char addrl=  (address & 0x00ff);      /*  7:0                 */
    unsigned char addrh= ((address & 0xff00) >> 8);/* 15:8                 */
    pic18f_core_instruction(k, 0x0E00 | addrl);    /* MOVLW <Addr>         */
    pic18f_core_instruction(k, 0x6EA9);            /* MOVWF EEADR          */
    pic18f_core_instruction(k, 0x0E00 | addrh);    /* MOVLW <AddrH>        */
    pic18f_core_instruction(k, 0x6EAA);            /* MOVWF EEADRH         */
}

/*
 * READ DATA MEMORY (EEPROM)
 */
unsigned char
pic18f_read_data_memory(struct k8048 *k)
{
    pic18f_core_instruction(k, 0x80A6);            /* BSF EECON1, RD       */
    pic18f_core_instruction(k, 0x50A8);            /* MOVF EEDATA, W, 0    */
    pic18f_core_instruction(k, 0x6EF5);            /* MOVWF TABLAT         */
    switch(pic18f_map[pic18f_index].datasheet)
    {
        case DS39752A:
        case DS39622K:
            pic18f_core_instruction(k, 0x0000);    /* NOP (DS39622K/39752A)*/
            break;
    }
    return pic18f_shift_out_tablat_register(k);
}

/*
 * WRITE DATA MEMORY (EEPROM)
 */
void
pic18f_write_data_memory(struct k8048 *k, unsigned char data)
{
    pic18f_core_instruction(k, 0x0E00 | data);     /* MOVLW <Data>         */
    pic18f_core_instruction(k, 0x6EA8);            /* MOVWF EEDATA         */
    pic18f_core_instruction(k, 0x84A6);            /* BSF EECON1, WREN     */
    switch(pic18f_map[pic18f_index].datasheet)
    {
         case DS39576B:
         case DS39592E: pic18f_core_instruction(k, 0x0E55);/* MOVLW 0x55   */
                        pic18f_core_instruction(k, 0x6EA7);/* MOVWF EECON2 */
                        pic18f_core_instruction(k, 0x0EAA);/* MOVLW 0xAA   */
                        pic18f_core_instruction(k, 0x6EA7);/* MOVWF EECON2 */
                        break;
    }
    pic18f_core_instruction(k, 0x82A6);            /* BSF EECON1, WR       */
    do  /* until write complete */
    {
        pic18f_core_instruction(k, 0x50A6);        /* MOVF EECON1, W, 0    */
        pic18f_core_instruction(k, 0x6EF5);        /* MOVWF TABLAT         */
        switch(pic18f_map[pic18f_index].datasheet)
        {
            case DS39752A:
            case DS39622K:
                pic18f_core_instruction(k, 0x0000);/* NOP (DS39622K/39752A)*/
                break;
        }
    }
    while( pic18f_shift_out_tablat_register(k) & 2 );
    pic18f_core_instruction(k, 0x94A6);            /* BCF EECON1, WREN     */
}   

/*****************************************************************************
 *
 * Programming Code/ID Panel
 *
 *****************************************************************************/

/*
 * INIT SINGLE PANEL WRITES
 *  
 *  DS39576B / DS39592E, NOT DS39622K
 */
void
pic18f_init_single_panel_writes(struct k8048 *k)
{ 
    pic18f_init_config_memory_access(k);           /* SWITCH TO CONFIG     */
    pic18f_set_table_pointer(k, 0x3C0006);         /* PANEL MODE CONFIG &  */
    pic18f_table_write(k, 0x0000);                 /* SINGLE PANEL WRITES  */
    pic18f_init_code_memory_access(k);             /* SWITCH TO CODE       */
}

/*
 * WRITE BUFFER (PANELSIZE BYTES INTO CODE PANEL, 8 BYTES INTO ID PANEL)
 */
void
pic18f_write_buffer(struct k8048 *k, unsigned long address, unsigned char buffer[MAXPANELSIZE])
{
    int i, limit;

    if( address == PIC18F_ID_LOW )                  /* ID   panel size is always 8 bytes */
        limit= 6;
    else                                            /* CODE panel size varies per device */
        limit= pic18f_map[pic18f_index].panelsize - 2;

    pic18f_set_table_pointer(k, address);

    for(i=0; i < limit; i+=2)
        pic18f_table_write_post_increment_2(k, (buffer[i+1] << 8) | buffer[i] );

    pic18f_table_write_start_programming(k, (buffer[i+1] << 8) | buffer[i] );

    pic18f_core_instruction_nopp(k);
}

/*
 * WRITE BUFFERED (SINGLE PANEL WRITES OF DEVICE PANEL SIZE)
 */
void
pic18f_write_buffered(struct k8048 *k, unsigned long address, unsigned char data, int mode)
{
    static unsigned char buffer[MAXPANELSIZE];
    static unsigned long buffer_address=0;
    static int write_pending=0;

    unsigned long new_address, boundary, mask;
    int i;

    switch(mode)
    {
        case WB_FLUSH:  if(write_pending)
                            pic18f_write_buffer(k, buffer_address, buffer);

        case WB_INIT:   for(i=0; i<pic18f_map[pic18f_index].panelsize; i++)
                            buffer[i]= WB_SPACE;

                        buffer_address=0;
                        write_pending =0;
                        break;

        case WB_ADD:    boundary= 0 - pic18f_map[pic18f_index].panelsize;
                        mask    = pic18f_map[pic18f_index].panelsize - 1;

                        new_address = address & boundary;
                        if(new_address != buffer_address)
                        {
                            if(write_pending)
                            {
                                pic18f_write_buffer(k, buffer_address, buffer);
                                for(i=0; i < pic18f_map[pic18f_index].panelsize; i++)
                                    buffer[i]= WB_SPACE;
                            }
                            buffer_address= new_address;
                        }

                        buffer[address & mask]= data;
                        write_pending= 1;
                        break;
    }
}

/*****************************************************************************
 *
 * Programming Config
 *
 *****************************************************************************/

/*
 * GOTO 100000
 */
void
pic18f_goto100000(struct k8048 *k)
{
    pic18f_core_instruction(k, 0xEF00);            /* GOTO 0x100000    */
    pic18f_core_instruction(k, 0xf800);            /* GOTO 0x100000    */
}

void
pic18f_write_configreg(struct k8048 *k, unsigned long address, int index)
{
    unsigned short data;

    data= pic18f_conf.index[ index ] | pic18f_conf.index[ index+1 ] << 8;

    pic18f_set_table_pointer(k, address);           /* must be an even address */
    pic18f_table_write_start_programming(k, data);
    pic18f_core_instruction_nopp(k);
    switch(pic18f_map[pic18f_index].datasheet)      /* increment address */
    {
        case DS39576B:                              /* INCF TBLPTRL */
        case DS39592E:  pic18f_core_instruction(k, 0x2AF6);
                        break;

        case DS39752A:                              /* TBLPTRL+=1   */
        case DS39622K:  pic18f_set_table_pointer(k, address + 1);
                        break;
    }
    pic18f_table_write_start_programming(k, data);
    pic18f_core_instruction_nopp(k);
}

/*
 * WRITE CONFIG (REGISTER 6H MUST BE WRITTEN LAST)
 */
void
pic18f_write_config(struct k8048 *k)
{
    pic18f_init_config_memory_access(k);
    pic18f_goto100000(k);
    pic18f_write_configreg(k, PIC18F_CONFIG_LOW +  0, PIC18F_CONFIG_1L);
    pic18f_write_configreg(k, PIC18F_CONFIG_LOW +  2, PIC18F_CONFIG_2L);
    pic18f_write_configreg(k, PIC18F_CONFIG_LOW +  4, PIC18F_CONFIG_3L);
    pic18f_write_configreg(k, PIC18F_CONFIG_LOW +  6, PIC18F_CONFIG_4L);
    pic18f_write_configreg(k, PIC18F_CONFIG_LOW +  8, PIC18F_CONFIG_5L);
    pic18f_write_configreg(k, PIC18F_CONFIG_LOW + 12, PIC18F_CONFIG_7L);
    pic18f_write_configreg(k, PIC18F_CONFIG_LOW + 10, PIC18F_CONFIG_6L);
}

/*****************************************************************************
 *
 * Programming
 *
 *****************************************************************************/

/*
 * DETERMINE MEMORY REGION: CODE, IDLOC, CONFIG or DATA
 */
int
pic18f_getregion(unsigned int address)
{
    if(address>=PIC18F_CODE_LOW && address<=PIC18F_ID_HIGH)
        return PIC18F_REGIONCODEID;

    if(address>=PIC18F_CONFIG_LOW && address<=PIC18F_CONFIG_HIGH)
        return PIC18F_REGIONCONFIG;

    if(address>=PIC18F_DATA_LOW && address<=PIC18F_DATA_HIGH)
        return PIC18F_REGIONDATA;

    printf("%s: warning: region unknown\n", __FUNCTION__);
    return PIC18F_REGIONUNKNOWN;
}

/*
 * INIT READ REGION
 */
void
pic18f_init_readregion(struct k8048 *k, int region)
{
    switch(region)
    {
        case PIC18F_REGIONCODEID:  
        case PIC18F_REGIONCONFIG:   pic18f_init_code_memory_access(k);
                                    break;
        case PIC18F_REGIONDATA:     pic18f_init_data_memory_access(k);
                                    break;
    }
}

/*
 * INIT WRITE REGION (CONFIG DELAYED)
 */
void
pic18f_init_writeregion(struct k8048 *k, int region)
{
    switch(region)
    {
        case PIC18F_REGIONCODEID:   switch(pic18f_map[pic18f_index].datasheet)
                                    {
                                        case DS39576B:
                                        case DS39592E:  pic18f_init_single_panel_writes(k);
                                                        break;
                                        default:        /* modern devices are ONLY single panel */
                                                        break;
                                    }
                                    pic18f_write_buffered(k, 0, 0, WB_INIT);
                                    break;
        case PIC18F_REGIONDATA:     pic18f_init_data_memory_access(k);
                                    break;
    }
}

/*
 * VERIFY REGION
 */
int
pic18f_verifyregion(struct k8048 *k, unsigned long address, int region, int index, unsigned char data)
{
    unsigned char vdata;

    switch(region)
    {
        case PIC18F_REGIONCODEID:   if(index==0)
                                        pic18f_set_table_pointer(k, address);
                                    vdata= pic18f_table_read_post_increment(k);
                                    break;
        case PIC18F_REGIONCONFIG:   /* can't realistically verify the config area assume okay */
                                    vdata=data;
                                    break;
        case PIC18F_REGIONDATA:     pic18f_set_data_pointer(k, address & PIC18F_DATA_MASK);
                                    vdata= pic18f_read_data_memory(k);
                                    break;
    }
    if(vdata != data)
    {
        printf("%s: error: read [%02x] expected [%02x] at [%04lx]\n", __FUNCTION__, vdata, data, address);
        return FAIL;
    }
    return PASS;
}

/*
 * WRITE REGION (CACHE CONFIG WORDS)
 */
void
pic18f_writeregion(struct k8048 *k, unsigned long address, int region, unsigned char data)
{
    switch(region)
    {
        case PIC18F_REGIONCODEID:   pic18f_write_buffered(k, address, data, WB_ADD);
                                    break;
        case PIC18F_REGIONCONFIG:   pic18f_conf.index[PIC18F_CONFIG_1L + (address & PIC18F_CONFIG_MASK)]= data;
                                    break;
        case PIC18F_REGIONDATA:     pic18f_set_data_pointer(k, address & PIC18F_DATA_MASK);
                                    pic18f_write_data_memory(k, data);
                                    break;
    }
}

/*
 * PROGRAM
 */
void
pic18f_program(struct k8048 *k)
{
    int i, j;
    unsigned long PC_address;
    int new_region, current_region=PIC18F_REGIONUNKNOWN;
    int total=0;

    io_init_program_verify(k);

    /* initialise device for programming */
    pic18f_bulk_erase(k);
    
    /* for each line */
    for(i=0; i<inhx32_count; i++)
    {
        PC_address= inhx32_pdata[i]->address;
        new_region= pic18f_getregion(PC_address);
        if(new_region==PIC18F_REGIONUNKNOWN)
            continue;

        if(new_region != current_region)
        {
            if(current_region==PIC18F_REGIONCODEID)
                pic18f_write_buffered(k, 0, 0, WB_FLUSH);
            pic18f_init_writeregion(k, new_region);
            current_region= new_region;
        }

        /* for each byte in line */
        for(j=0; j<inhx32_pdata[i]->nbytes; j++)
        {
            pic18f_writeregion(k, PC_address++, current_region, inhx32_pdata[i]->bytes[j]);
            total++;
        }
    }
    if(current_region==PIC18F_REGIONCODEID)
        pic18f_write_buffered(k, 0, 0, WB_FLUSH);
        
    io_standby(k);

    /* finalise device programming (write config words) */
    pic18f_program_config(k);
    total+=14;

    printf("Total: %d\n", total);
}

/*
 * VERIFY (THIS WILL FAIL IF PROTECTION WAS ENABLED EARLIER)
 */
int
pic18f_verify(struct k8048 *k)
{
    int i, j;
    unsigned long PC_address;
    int new_region, current_region=PIC18F_REGIONUNKNOWN;
    int fail=0, total=0;

    io_init_program_verify(k);

    /* for each line */
    for(i=0; i<inhx32_count; i++)
    {
        PC_address= inhx32_pdata[i]->address;
        new_region= pic18f_getregion(PC_address);
        if(new_region==PIC18F_REGIONUNKNOWN)
            continue;

        if(new_region != current_region)
        {
            pic18f_init_readregion(k, new_region);
            current_region= new_region;
        }

        /* for each byte in line */
        for(j=0; j<inhx32_pdata[i]->nbytes; j+=1)
        {
            fail+=pic18f_verifyregion(k, PC_address++, current_region, j, inhx32_pdata[i]->bytes[j]);
            total++;
        }
    }
    io_standby(k);

    printf("Total: %d Pass: %d Fail: %d\n", total, total-fail, fail);
    return fail;
}
 
/*
 * PROGRAM CONFIGURATION (AFTER VERIFY)
 */
void
pic18f_program_config(struct k8048 *k)
{
    io_init_program_verify(k);

    pic18f_write_config(k);

    io_standby(k);
}

/*****************************************************************************
 *
 * Diagnostic functions
 *
 *****************************************************************************/

/*
 * DUMP DEVICE ID DETAILS
 */
void
pic18f_dumpdeviceid(struct k8048 *k)
{
    unsigned short deviceid, revision;

    deviceid= (pic18f_conf.index[PIC18F_CONFIG_DEVICEID2]<<8) | (pic18f_conf.index[PIC18F_CONFIG_DEVICEID1] & PIC18F_DEVICEID1_IDMASK);
    revision=  pic18f_conf.index[PIC18F_CONFIG_DEVICEID1] & PIC18F_DEVICEID1_REVMASK;

    printf("[200000] [IDLOCATION1] %02x\n", pic18f_conf.index[PIC18F_CONFIG_IDLOCATION1]);
    printf("[200001] [IDLOCATION2] %02x\n", pic18f_conf.index[PIC18F_CONFIG_IDLOCATION2]);
    printf("[200002] [IDLOCATION3] %02x\n", pic18f_conf.index[PIC18F_CONFIG_IDLOCATION3]);
    printf("[200003] [IDLOCATION4] %02x\n", pic18f_conf.index[PIC18F_CONFIG_IDLOCATION4]);
    printf("[200004] [IDLOCATION5] %02x\n", pic18f_conf.index[PIC18F_CONFIG_IDLOCATION5]);
    printf("[200005] [IDLOCATION6] %02x\n", pic18f_conf.index[PIC18F_CONFIG_IDLOCATION6]);
    printf("[200006] [IDLOCATION7] %02x\n", pic18f_conf.index[PIC18F_CONFIG_IDLOCATION7]);
    printf("[200007] [IDLOCATION8] %02x\n", pic18f_conf.index[PIC18F_CONFIG_IDLOCATION8]);
    printf("[3FFFFE] [DEVICEID]    %04x DEV:%x (%x:%x) REV:%x",
        /* literal */(pic18f_conf.index[PIC18F_CONFIG_DEVICEID2] << 8) | pic18f_conf.index[PIC18F_CONFIG_DEVICEID1],
        /* actual  */deviceid,
        /* high    */(deviceid >> 8),
        /* low     */(deviceid >> 5) & 7,
        revision);

    printf(" %s\n", pic18f_map[pic18f_index].devicename);
}

/*
 * DUMP CONFIG WORD DETAILS FOR DEVICE (INCOMPLETE MAP)
 */
void
pic18f_dumpconfig(struct k8048 *k)
{
    unsigned short deviceid;
    unsigned char w;

    deviceid= (pic18f_conf.index[PIC18F_CONFIG_DEVICEID2]<<8) | (pic18f_conf.index[PIC18F_CONFIG_DEVICEID1] & PIC18F_DEVICEID1_IDMASK);

    printf("[300000] [CONFIG1] %04x\n", (pic18f_conf.index[PIC18F_CONFIG_1H]<<8) | pic18f_conf.index[PIC18F_CONFIG_1L]);
    printf("[300002] [CONFIG2] %04x\n", (pic18f_conf.index[PIC18F_CONFIG_2H]<<8) | pic18f_conf.index[PIC18F_CONFIG_2L]);
    printf("[300004] [CONFIG3] %04x\n", (pic18f_conf.index[PIC18F_CONFIG_3H]<<8) | pic18f_conf.index[PIC18F_CONFIG_3L]);
    printf("[300006] [CONFIG4] %04x\n", (pic18f_conf.index[PIC18F_CONFIG_4H]<<8) | pic18f_conf.index[PIC18F_CONFIG_4L]);
    printf("[300008] [CONFIG5] %04x\n", (pic18f_conf.index[PIC18F_CONFIG_5H]<<8) | pic18f_conf.index[PIC18F_CONFIG_5L]);
    printf("[30000A] [CONFIG6] %04x\n", (pic18f_conf.index[PIC18F_CONFIG_6H]<<8) | pic18f_conf.index[PIC18F_CONFIG_6L]);
    printf("[30000C] [CONFIG7] %04x\n", (pic18f_conf.index[PIC18F_CONFIG_7H]<<8) | pic18f_conf.index[PIC18F_CONFIG_7L]);

    printf("\n[CONFIG_1H]\n");
    w= pic18f_conf.index[PIC18F_CONFIG_1H] & 0x20;
    if(w)
        printf("\t[OSCCEN] Disabled\n");
    else
        printf("\t[OSCCEN] Enabled\n");

    /*
     * XXX these values vary per device type
     */
    w= pic18f_conf.index[PIC18F_CONFIG_1H] & 0x0f;
    switch(w)
    {
        case 15:/* PIC18F1320 */
        case 14:
        case 13:
        case 12:printf("\t[FOSC3:FOSC0] 11XX EXT RC CLKOUT\n");break;
        case 11:printf("\t[FOSC3:FOSC0] 1011 UNKNOWN\n");break;
        case 10:printf("\t[FOSC3:FOSC0] 1010 UNKNOWN\n");break;

        case 9: printf("\t[FOSC3:FOSC0] 1001 INT RC CLKOUT\n"); break;
        case 8: printf("\t[FOSC3:FOSC0] 1000 INT RC NOCLKOUT\n"); break;
        case 7: printf("\t[FOSC3:FOSC0] 0111 EXT RC\n"); break;
        case 6: printf("\t[FOSC3:FOSC0] 0110 HS\n"); break;
        case 5: printf("\t[FOSC3:FOSC0] 0101 EC\n"); break;
        case 4: printf("\t[FOSC3:FOSC0] 0100 EC CLKOUT\n"); break;
        case 3: printf("\t[FOSC3:FOSC0] 0011 RC\n"); break;
        case 2: printf("\t[FOSC3:FOSC0] 0010 HS\n"); break;
        case 1: printf("\t[FOSC3:FOSC0] 0001 XT\n"); break;
        case 0: printf("\t[FOSC3:FOSC0] 0000 LP\n"); break;
    }
    
    printf("\n[CONFIG_2L]\n");
    w= (pic18f_conf.index[PIC18F_CONFIG_2L] & 0x0C) >> 2;
    switch(w)
    {
        case 3: printf("\t[BORV1:BORV0] 2.5V\n"); break;
        case 2: printf("\t[BORV1:BORV0] 2.7V\n"); break;
        case 1: printf("\t[BORV1:BORV0] 4.2V\n"); break;
        case 0: printf("\t[BORV1:BORV0] 4.5V\n"); break;
    }
    
    w= pic18f_conf.index[PIC18F_CONFIG_2L] & 0x02;
    if(w)
        printf("\t[BOREN] Enabled\n");
    else
        printf("\t[BOREN] Disabled\n");

    w= pic18f_conf.index[PIC18F_CONFIG_2L] & 0x01;
    if(w)
        printf("\t[PWRTEN] Disabled\n");
    else
        printf("\t[PWRTEN] Enabled\n");
    
    printf("\n[CONFIG_2H]\n");
    w= (pic18f_conf.index[PIC18F_CONFIG_2H] & 0x0E) >> 1;
    switch(w)
    {
        case 7: printf("\t[WDTP2:WDTP0] 1:128\n"); break;
        case 6: printf("\t[WDTP2:WDTP0] 1:64\n");  break;
        case 5: printf("\t[WDTP2:WDTP0] 1:32\n");  break;
        case 4: printf("\t[WDTP2:WDTP0] 1:16\n");  break;
        case 3: printf("\t[WDTP2:WDTP0] 1:8\n");   break;
        case 2: printf("\t[WDTP2:WDTP0] 1:4\n");   break;
        case 1: printf("\t[WDTP2:WDTP0] 1:2\n");   break;
        case 0: printf("\t[WDTP2:WDTP0] 1:1\n");   break;
    }
    
    w= pic18f_conf.index[PIC18F_CONFIG_2H] & 0x01;
    if(w)
        printf("\t[WDTEN] Enabled\n");
    else
        printf("\t[WDTEN] Disabled\n");
    
    printf("\n[CONFIG_3H]\n");
    w= pic18f_conf.index[PIC18F_CONFIG_3H] & 0x01;
    if(w)
        printf("\t[CCP2MX] RC1\n");
    else
        printf("\t[CCP2MX] RB3\n");
    
    printf("\n[CONFIG_4L]\n");
    w= pic18f_conf.index[PIC18F_CONFIG_4L] & 0x80;
    if(w)
        printf("\t[DEBUG] Disabled\n");
    else
        printf("\t[DEBUG] Enabled\n");
    
    w= pic18f_conf.index[PIC18F_CONFIG_4L] & 0x04;
    if(w)
        printf("\t[LVP] Enabled\n");
    else
        printf("\t[LVP] Disabled\n");
    
    w= pic18f_conf.index[PIC18F_CONFIG_4L] & 0x01;
    if(w)
        printf("\t[STVREN] Enabled\n");
    else
        printf("\t[STVREN] Disabled\n");

    /*
     * PROTECTION BITS
     */
    
    printf("\n[CONFIG_5L]");
    w= pic18f_conf.index[PIC18F_CONFIG_5L] & 0x08;
    if(w)
        printf("\t[CP3]   Off");
    else
        printf("\t[CP3]   On");
    w= pic18f_conf.index[PIC18F_CONFIG_5L] & 0x04;
    if(w)
        printf("\t[CP2]   Off");
    else
        printf("\t[CP2]   On");
    w= pic18f_conf.index[PIC18F_CONFIG_5L] & 0x02;
    if(w)
        printf("\t[CP1]   Off");
    else
        printf("\t[CP1]   On");
    w= pic18f_conf.index[PIC18F_CONFIG_5L] & 0x01;
    if(w)
        printf("\t[CP0]   Off");
    else
        printf("\t[CP0]   On");

    printf("\n[CONFIG_5H]");
    w= pic18f_conf.index[PIC18F_CONFIG_5H] & 0x80;
    if(w)
        printf("\t[CPD]   Off");
    else
        printf("\t[CPD]   On");
    w= pic18f_conf.index[PIC18F_CONFIG_5H] & 0x40;
    if(w)
        printf("\t[CPB]   Off");
    else
        printf("\t[CPB]   On");

    printf("\n[CONFIG_6L]");
    w= pic18f_conf.index[PIC18F_CONFIG_6L] & 0x08;
    if(w)
        printf("\t[WRT3]  Off");
    else
        printf("\t[WRT3]  On");
    w= pic18f_conf.index[PIC18F_CONFIG_6L] & 0x04;
    if(w)
        printf("\t[WRT2]  Off");
    else
        printf("\t[WRT2]  On");
    w= pic18f_conf.index[PIC18F_CONFIG_6L] & 0x02;
    if(w)
        printf("\t[WRT1]  Off");
    else
        printf("\t[WRT1]  On");
    w= pic18f_conf.index[PIC18F_CONFIG_6L] & 0x01;
    if(w)
        printf("\t[WRT0]  Off");
    else
        printf("\t[WRT0]  On");
    
    printf("\n[CONFIG_6H]");
    w= pic18f_conf.index[PIC18F_CONFIG_6H] & 0x80;
    if(w)
        printf("\t[WRTD]  Off");
    else
        printf("\t[WRTD]  On");
    w= pic18f_conf.index[PIC18F_CONFIG_6H] & 0x40;
    if(w)
        printf("\t[WRTB]  Off");
    else
        printf("\t[WRTB]  On");
    w= pic18f_conf.index[PIC18F_CONFIG_6H] & 0x20;
    if(w)
        printf("\t[WRTC]  Off");
    else
        printf("\t[WRTC]  On");
    
    printf("\n[CONFIG_7L]");
    w= pic18f_conf.index[PIC18F_CONFIG_7L] & 0x08;
    if(w)
        printf("\t[EBTR3] Off");
    else
        printf("\t[EBTR3] On");
    w= pic18f_conf.index[PIC18F_CONFIG_7L] & 0x04;
    if(w)
        printf("\t[EBTR2] Off");
    else
        printf("\t[EBTR2] On");
    w= pic18f_conf.index[PIC18F_CONFIG_7L] & 0x02;
    if(w)
        printf("\t[EBTR1] Off");
    else
        printf("\t[EBTR1] On");
    w= pic18f_conf.index[PIC18F_CONFIG_7L] & 0x01;
    if(w)
        printf("\t[EBTR0] Off");
    else
        printf("\t[EBTR0] On");
    
    printf("\n[CONFIG_7H]");
    w= pic18f_conf.index[PIC18F_CONFIG_7H] & 0x40;
    if(w)
        printf("\t[EBTRB] Off\n");
    else
        printf("\t[EBTRB] On\n");
}
